<?php
/**
 * Per-product shipping product editor setup class file.
 *
 * @package WC_Shipping_Per_Product
 */

namespace WooCommerce\Shipping\PerProduct;

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

use WC_Shipping_Per_Product_Admin;
use WC_Shipping_Per_Product_Init;
use Automattic\WooCommerce\Admin\Features\ProductBlockEditor\BlockRegistry;
use Automattic\WooCommerce\Admin\BlockTemplates\BlockInterface;
use WP_REST_Request;
use WC_Product;

/**
 * Per-product shipping product editor setup class.
 *
 * You can extend this class with
 * WP_REST_Controller / WC_REST_Controller / WC_REST_Products_V2_Controller / WC_REST_CRUD_Controller etc.
 * Found in packages/woocommerce-rest-api/src/Controllers/
 *
 * @since 3.0.0
 */
class Product_Editor_Setup {
	/**
	 * Plugin main class instance.
	 *
	 * @var WC_Shipping_Per_Product_Init
	 */
	protected WC_Shipping_Per_Product_Init $per_product;

	/**
	 * Constructor.
	 *
	 * @param WC_Shipping_Per_Product_Init $per_product Referance to init object.
	 *
	 * @return void
	 */
	public function __construct( WC_Shipping_Per_Product_Init $per_product ) {
		$this->per_product = $per_product;

		add_action( 'init', array( $this, 'extension_block_init' ) );
		add_action( 'woocommerce_block_template_area_product-form_after_add_block_product-fee-and-dimensions-section', array( $this, 'add_blocks' ) );
		add_action( 'woocommerce_block_template_area_product-form_after_add_block_product-variation-fee-and-dimensions-section', array( $this, 'add_blocks' ) );
		add_filter( 'woocommerce_rest_api_get_rest_namespaces', array( $this, 'add_per_product_rest_api_endpoint' ) );
		add_filter( 'woocommerce_rest_pre_insert_product_object', array( $this, 'add_per_product_data_to_request' ), 10, 3 );
		add_filter( 'woocommerce_rest_pre_insert_product_variation_object', array( $this, 'add_per_product_data_to_request' ), 10, 3 );
	}

	/**
	 * Registers the block using the metadata loaded from the `block.json` file.
	 * Behind the scenes, it registers also all assets so they can be enqueued
	 * through the block editor in the corresponding context.
	 *
	 * @see https://developer.wordpress.org/reference/functions/register_block_type/
	 *
	 * @return void
	 */
	public function extension_block_init() {
		// phpcs:ignore WordPress.Security.NonceVerification.Recommended --- No need to use nonce as no DB operation required.
		if ( isset( $_GET['page'] ) && 'wc-admin' === $_GET['page'] ) {
			BlockRegistry::get_instance()->register_block_type_from_metadata( PER_PRODUCT_SHIPPING_ABSPATH . 'dist/product-editor' );
		}
	}

	/**
	 * Add "per product shipping" blocks into new product editor.
	 *
	 * @param BlockInterface $product_section Product section block.
	 *
	 * @return void
	 */
	public function add_blocks( BlockInterface $product_section ) {
		$parent = $product_section->get_parent();

		if ( ! method_exists( $parent, 'add_section' ) ) {
			return;
		}

		// Parent Section.
		$per_product_section = $parent->add_section(
			array(
				'id'         => 'wc-shipping-per-product-section',
				'order'      => 30,
				'attributes' => array(
					'title'       => __( 'Per-product shipping', 'woocommerce-shipping-per-product' ),
					'description' => __( 'Configure per-product shipping settings for this product.', 'woocommerce-shipping-per-product' ),
				),
			)
		);

		$per_product_section->add_block(
			array(
				'id'         => WC_Shipping_Per_Product_Admin::META_KEY_ENABLE_PP,
				'order'      => 10,
				'blockName'  => 'woocommerce/product-checkbox-field',
				'attributes' => array(
					'property'       => 'meta_data.' . WC_Shipping_Per_Product_Admin::META_KEY_ENABLE_PP,
					'label'          => __( 'Enable per-product shipping cost', 'woocommerce-shipping-per-product' ),
					'checkedValue'   => 'yes',
					'uncheckedValue' => '',
				),
			)
		);

		if ( false !== $this->per_product->use_legacy_shipping_method() ) {
			$per_product_section->add_block(
				array(
					'id'             => WC_Shipping_Per_Product_Admin::META_KEY_ADD_TO_ALL,
					'order'          => 10,
					'blockName'      => 'woocommerce/product-checkbox-field',
					'attributes'     => array(
						'property'       => 'meta_data.' . WC_Shipping_Per_Product_Admin::META_KEY_ADD_TO_ALL,
						'label'          => __( 'Add per-product shipping cost to all shipping method rates?', 'woocommerce-shipping-per-product' ),
						'checkedValue'   => 'yes',
						'uncheckedValue' => 'no',
					),
					'hideConditions' => array(
						array(
							'expression' => 'editedProduct.meta_data.' . WC_Shipping_Per_Product_Admin::META_KEY_ENABLE_PP . ' !== "yes"',
						),
					),
				)
			);
		}

		$per_product_section->add_block(
			array(
				'id'             => 'extension-wc-shipping-per-product',
				'order'          => 20,
				'blockName'      => 'extension/wc-shipping-per-product',
				'attributes'     => array(
					'table_text'        => WC_Shipping_Per_Product_Admin::get_i18n_table_text(),
					'table_urls'        => WC_Shipping_Per_Product_Admin::get_table_urls(),
					'rule_count_limit'  => WC_Shipping_Per_Product_Admin::$rule_count_limit,
					'decimal_separator' => wc_get_price_decimal_separator(),
				),
				'hideConditions' => array(
					array(
						'expression' => 'editedProduct.meta_data.' . WC_Shipping_Per_Product_Admin::META_KEY_ENABLE_PP . ' !== "yes"',
					),
				),
			)
		);
	}

	/**
	 * Add plugin custom WC REST API endpoint.
	 *
	 * @param array $controllers List of WooCommerce controllers.
	 *
	 * @return array
	 */
	public function add_per_product_rest_api_endpoint( array $controllers ): array {
		$controllers['wc/v3']['per_product'] = 'WooCommerce\Shipping\PerProduct\REST_Controller';

		return $controllers;
	}

	/**
	 * Add Per-Product data to REST request.
	 *
	 * @param WC_Product      $product  Product object.
	 * @param WP_REST_Request $request  REST request.
	 * @param boolean         $creating If is creating a new object.
	 *
	 * @return WC_Product
	 */
	public function add_per_product_data_to_request( WC_Product $product, WP_REST_Request $request, bool $creating ): WC_Product {
		$meta_data = $request->get_param( 'meta_data' );

		if ( empty( $meta_data ) ) {
			return $product;
		}

		$filtered_rules_meta = array_filter(
			$meta_data,
			function( $meta ) {
				return 'per_product_rules' === $meta['key'];
			}
		);

		if ( empty( $filtered_rules_meta ) ) {
			return $product;
		}

		$rules = array_shift( $filtered_rules_meta );

		if ( empty( $rules['value'] ) ) {
			return $product;
		}

		$filtered_enable_pp_meta = array_filter(
			$meta_data,
			function( $meta ) {
				return '_per_product_shipping' === $meta['key'];
			}
		);

		$enable_pp = ! empty( $filtered_enable_pp_meta ) ? array_shift( $filtered_enable_pp_meta ) : array( 'value' => '' );

		$enabled              = ! empty( $enable_pp['value'] );
		$saved_enable_setting = 'yes' === $product->get_meta( '_per_product_shipping' );

		WC_Shipping_Per_Product_Admin::clear_shipping_cache_if_necessary( $enabled, $saved_enable_setting );

		$formatted_rules = WC_Shipping_Per_Product_Admin::formatting_rules_data( $rules['value'] );
		WC_Shipping_Per_Product_Admin::save_product_rules( $product->get_id(), $formatted_rules );

		$product->update_meta_data( 'per_product_rules', WC_Shipping_Per_Product_Admin::get_rest_formatted_rules( $product->get_id() ) );
		$product->save_meta_data();

		return $product;
	}
}
