<?php
/**
 * Per Product Shipping Custom REST class file.
 *
 * @package WC_Shipping_Per_Product
 */

namespace WooCommerce\Shipping\PerProduct;

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

use WC_Product;
use WC_Shipping_Per_Product_Admin;
use WP_REST_Server;
use WP_REST_Request;

/**
 * Per Product Shipping Custom REST class.
 *
 * You can extend this class with
 * WP_REST_Controller / WC_REST_Controller / WC_REST_Products_V2_Controller / WC_REST_CRUD_Controller etc.
 * Found in packages/woocommerce-rest-api/src/Controllers/
 *
 * @since 3.0.0
 */
class REST_Controller {
	/**
	 * WooCommerce REST API namespace.
	 *
	 * @var string
	 */
	protected string $namespace = 'wc/v3';

	/**
	 * This plugin REST base.
	 *
	 * @var string
	 */
	protected string $rest_base = 'per_product/(?P<id>\d+)';

	/**
	 * Get custom data.
	 *
	 * @param WP_REST_Request $data Request data.
	 *
	 * @return boolean|array.
	 */
	public function get_per_product_rules( WP_REST_Request $data ) {

		$product = wc_get_product( $data->get_param( 'id' ) );

		if ( ! $product instanceof WC_Product ) {
			return false;
		}

		if ( $product->get_id() < 1 ) {
			return false;
		}

		$rules_with_index = WC_Shipping_Per_Product_Admin::get_rest_formatted_rules( $product->get_id() );

		return array(
			'loaded_flag' => 'Data Rule ' . $product->get_id() . ' is loaded',
			'rules'       => $rules_with_index,
		);
	}

	/**
	 * Registering the custom routes.
	 *
	 * @return void.
	 */
	public function register_routes() {
		register_rest_route(
			$this->namespace,
			'/' . $this->rest_base,
			array(
				'methods'             => WP_REST_Server::READABLE,
				'callback'            => array( $this, 'get_per_product_rules' ),
				'permission_callback' => array( $this, 'get_per_product_rules_permissions_check' ),
			)
		);
	}

	/**
	 * Permission check for the REST API route.
	 *
	 * @return WP_Error|boolean
	 */
	public function get_per_product_rules_permissions_check() {
		// phpcs:ignore --- WooCommerce custom capability.
		if ( ! current_user_can( 'edit_products' ) ) {
			return new \WP_Error(
				'woocommerce_rest_cannot_edit',
				__( 'Sorry, you cannot read this resource.', 'woocommerce-shipping-per-product' ),
				array(
					'status' => rest_authorization_required_code(),
				)
			);
		}

		return true;
	}
}
