<?php
/**
 * Admin handler class.
 *
 * @package WC_Shipping_Per_Product
 */

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

use WooCommerce\Shipping\PerProduct\Product_Editor_Setup;

/**
 * Admin handler.
 */
class WC_Shipping_Per_Product_Admin {

	public const META_KEY_ENABLE_PP  = '_per_product_shipping';
	public const META_KEY_ENABLE_PV  = '_per_variation_shipping';
	public const META_KEY_ADD_TO_ALL = '_per_product_shipping_add_to_all';

	/**
	 * Plugin main class instance.
	 *
	 * @var WC_Shipping_Per_Product_Init
	 */
	protected $per_product;
	/**
	 * The maximum number of rules to display in the product editor.
	 * If a product has more rules than this, the rules will need to
	 * be managed offsite and imported (overridden).
	 *
	 * @var int
	 */
	public static $rule_count_limit = 200;

	/**
	 * Constructor.
	 *
	 * @param WC_Shipping_Per_Product_Init $per_product Referance to init object.
	 */
	public function __construct( WC_Shipping_Per_Product_Init $per_product ) {
		$this->per_product = $per_product;

		if ( is_admin() ) {
			add_action( 'admin_enqueue_scripts', array( $this, 'admin_enqueue_scripts' ) );
			add_action( 'woocommerce_product_options_shipping', array( $this, 'product_options' ) );
			add_action( 'woocommerce_variation_options', array( $this, 'variation_options' ), 10, 3 );
			add_action( 'woocommerce_product_after_variable_attributes', array( $this, 'product_after_variable_attributes' ), 10, 3 );
			add_action( 'woocommerce_process_product_meta', array( $this, 'save' ) );
			add_action( 'woocommerce_save_product_variation', array( $this, 'save_variation' ), 10, 2 );
			add_action( 'woocommerce_product_duplicate', array( $this, 'duplicate_rules' ), 10, 2 );
			add_action( 'wp_ajax_wc_shipping_per_product_export_rules', array( $this, 'export_rules' ) );
		}

		// Setting up the new product editor user interface.
		new Product_Editor_Setup( $this->per_product );
	}

	/**
	 * Scripts and styles.
	 */
	public function admin_enqueue_scripts() {
		$suffix = defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ? '' : '.min';

		wp_enqueue_style( 'wc-shipping-per-product-styles', plugins_url( 'assets/css/admin.css', PER_PRODUCT_SHIPPING_FILE ), array(), PER_PRODUCT_SHIPPING_VERSION );
		wp_register_script( 'wc-shipping-per-product', plugins_url( 'assets/js/shipping-per-product' . $suffix . '.js', PER_PRODUCT_SHIPPING_FILE ), array( 'jquery' ), PER_PRODUCT_SHIPPING_VERSION, true );

		wp_localize_script(
			'wc-shipping-per-product',
			'wc_shipping_per_product_params',
			array(
				'i18n_no_row_selected' => __( 'No row selected', 'woocommerce-shipping-per-product' ),
				'i18n_product_id'      => __( 'Product ID', 'woocommerce-shipping-per-product' ),
				'i18n_country_code'    => __( 'Country Code', 'woocommerce-shipping-per-product' ),
				'i18n_state'           => __( 'State/County Code', 'woocommerce-shipping-per-product' ),
				'i18n_postcode'        => __( 'Zip/Postal Code', 'woocommerce-shipping-per-product' ),
				'i18n_cost'            => __( 'Cost', 'woocommerce-shipping-per-product' ),
				'i18n_item_cost'       => __( 'Item Cost', 'woocommerce-shipping-per-product' ),
				'decimal_separator'    => wc_get_price_decimal_separator(),
			)
		);

		wp_localize_script(
			'wc-shipping-per-product',
			'wc_shipping_per_product_errors',
			array(
				'i18n_incorrect_country_code' => __( 'The country code does not exist.', 'woocommerce-shipping-per-product' ),
			)
		);

		wp_localize_script( 'wc-shipping-per-product', 'wc_shipping_per_product_countries', WC()->countries->get_countries() );
	}

	/**
	 * List of texts being used on the per product shipping table.
	 *
	 * @return array
	 */
	public static function get_i18n_table_text() {
		return array(
			'i18n_country_code'        => __( 'Country Code', 'woocommerce-shipping-per-product' ),
			'i18n_country_code_tip'    => __( 'A 2 digit country code, e.g. US. Leave blank to apply to all.', 'woocommerce-shipping-per-product' ),
			'i18n_state_code'          => __( 'State/County Code', 'woocommerce-shipping-per-product' ),
			'i18n_state_code_tip'      => __( 'A state code, e.g. AL. Leave blank to apply to all.', 'woocommerce-shipping-per-product' ),
			'i18n_zip_code'            => __( 'Zip/Postal Code', 'woocommerce-shipping-per-product' ),
			'i18n_zip_code_tip'        => __( 'Postcode for this rule. Wildcards (*) can be used. Leave blank to apply to all areas.', 'woocommerce-shipping-per-product' ),
			'i18n_line_cost'           => __( 'Line Cost (Excl. Tax)', 'woocommerce-shipping-per-product' ),
			'i18n_line_cost_tip'       => __( 'Decimal cost for the line as a whole.', 'woocommerce-shipping-per-product' ),
			'i18n_item_cost'           => __( 'Item Cost (Excl. Tax)', 'woocommerce-shipping-per-product' ),
			'i18n_item_cost_tip'       => __( 'Decimal cost for the item (multiplied by qty).', 'woocommerce-shipping-per-product' ),
			'i18n_helper_text'         => __( 'View helper text', 'woocommerce-shipping-per-product' ),
			'i18n_insert_row'          => __( 'Insert row', 'woocommerce-shipping-per-product' ),
			'i18n_remove_row'          => __( 'Remove row', 'woocommerce-shipping-per-product' ),
			'i18n_export_csv'          => __( 'Export CSV', 'woocommerce-shipping-per-product' ),
			'i18n_import_csv'          => __( 'Import CSV', 'woocommerce-shipping-per-product' ),
			'i18n_import_csv_override' => __( 'Import CSV (override)', 'woocommerce-shipping-per-product' ),
			'i18n_override_conf_text'  => __( 'This will remove all existing shipping rules and replace them with the rule(s) in your imported CSV file. Are you sure this is what you intend to do?', 'woocommerce-shipping-per-product' ),
			'i18n_csv_product_id'      => __( 'Product ID', 'woocommerce-shipping-per-product' ),
			'i18n_csv_country_code'    => __( 'Country Code', 'woocommerce-shipping-per-product' ),
			'i18n_csv_state'           => __( 'State/County Code', 'woocommerce-shipping-per-product' ),
			'i18n_csv_postcode'        => __( 'Zip/Postal Code', 'woocommerce-shipping-per-product' ),
			'i18n_csv_cost'            => __( 'Cost', 'woocommerce-shipping-per-product' ),
			'i18n_csv_item_cost'       => __( 'Item Cost', 'woocommerce-shipping-per-product' ),
			'i18n_loading_rule'        => __( 'Loading the rule...', 'woocommerce-shipping-per-product' ),
			'i18n_saving_rule'         => __( 'Saving the rule...', 'woocommerce-shipping-per-product' ),
			'i18n_empty_rule'          => __( 'There is no rules yet.', 'woocommerce-shipping-per-product' ),
		);
	}

	/**
	 * Get table URLs.
	 *
	 * @return array.
	 */
	public static function get_table_urls() {
		$override_nonce = wp_create_nonce( 'override-product-id' );
		$import_url     = add_query_arg(
			array(
				'import' => 'woocommerce_per_product_shipping_csv',
			),
			admin_url( 'admin.php' )
		);
		$override_url   = add_query_arg(
			array(
				'_wpnonce_override-product-id' => $override_nonce,
				'override_product_id'          => '%s',
			),
			$import_url
		);

		return array(
			'import_url'          => $import_url,
			'import_override_url' => $override_url,
		);
	}

	/**
	 * Output product options.
	 */
	public function product_options() {
		global $post, $wpdb;

		wp_enqueue_script( 'wc-shipping-per-product' );

		echo '</div><div class="options_group per_product_shipping">';

		woocommerce_wp_checkbox(
			array(
				'id'          => self::META_KEY_ENABLE_PP,
				'label'       => __( 'Per-product shipping', 'woocommerce-shipping-per-product' ),
				'description' => __( 'Enable per-product shipping cost', 'woocommerce-shipping-per-product' ),
			)
		);

		$this->output_rules();
	}

	/**
	 * Output variation options.
	 *
	 * @param int     $loop           Loop index.
	 * @param array   $variation_data Variation data.
	 * @param WP_Post $variation      Post instance of variation.
	 */
	public function variation_options( $loop, $variation_data, $variation ) {
		wp_enqueue_script( 'wc-shipping-per-product' );
		?>
		<label><input type="checkbox" class="checkbox enable_per_product_shipping" name="<?php echo esc_attr( self::META_KEY_ENABLE_PV . '[' . $variation->ID . ']' ); ?>" <?php checked( get_post_meta( $variation->ID, self::META_KEY_ENABLE_PP, true ), 'yes' ); ?> /> <?php esc_html_e( 'Per-variation shipping', 'woocommerce-shipping-per-product' ); ?></label>
		<?php
	}

	/**
	 * Show Rules.
	 *
	 * @param int     $loop           Loop index.
	 * @param array   $variation_data Variation data.
	 * @param WP_Post $variation      Post instance of variation.
	 */
	public function product_after_variable_attributes( $loop, $variation_data, $variation ) {
		echo '<tr class="per_product_shipping per_variation_shipping"><td colspan="2">';
		$this->output_rules( $variation->ID );
		echo '</td></tr>';
	}

	/**
	 * Get the count of shipping rules of certain product.
	 *
	 * @param int $product_id Product ID.
	 */
	public static function get_shipping_rules_count( $product_id ) {
		global $wpdb;

		return $wpdb->get_var( $wpdb->prepare( "SELECT COUNT(*) FROM {$wpdb->prefix}woocommerce_per_product_shipping_rules WHERE product_id = %d ORDER BY rule_order;", $product_id ) );
	}

	/**
	 * Get all shipping rules for certain product.
	 *
	 * @param int $product_id Product ID.
	 */
	public static function get_shipping_rules( $product_id ) {
		global $wpdb;

		return $wpdb->get_results( $wpdb->prepare( "SELECT * FROM {$wpdb->prefix}woocommerce_per_product_shipping_rules WHERE product_id = %d ORDER BY rule_order;", $product_id ) );
	}

	/**
	 * Output rules table.
	 *
	 * @param int $post_id Post ID.
	 */
	public function output_rules( $post_id = 0 ) {
		global $post, $wpdb;

		if ( ! $post_id ) {
			$post_id = $post->ID;
		}

		$table_text       = self::get_i18n_table_text();
		$table_urls       = self::get_table_urls();
		$count            = self::get_shipping_rules_count( $post_id );
		$cost_placeholder = sprintf( '0%1$s00', wc_get_price_decimal_separator() );
		?>
		<div class="rules per_product_shipping_rules">

			<?php
			if ( false !== $this->per_product->use_legacy_shipping_method() ) {
				woocommerce_wp_checkbox(
					array(
						'id'          => self::META_KEY_ADD_TO_ALL . '[' . $post_id . ']',
						'label'       => __( 'Adjust Shipping Costs', 'woocommerce-shipping-per-product' ),
						'description' => __( 'Add per-product shipping cost to all shipping method rates?', 'woocommerce-shipping-per-product' ),
						'value'       => get_post_meta( $post_id, self::META_KEY_ADD_TO_ALL, true ),
					)
				);
			}

			wp_nonce_field( 'per-product-export', '_export_nonce' );
			?>

			<table class="widefat">
				<thead>
					<tr>
						<th>&nbsp;</th>
						<th><?php echo esc_html( $table_text['i18n_country_code'] ); ?>&nbsp;<a class="tips" data-tip="<?php echo esc_attr( $table_text['i18n_country_code_tip'] ); ?>">[?]</a></th>
						<th><?php echo esc_html( $table_text['i18n_state_code'] ); ?>&nbsp;<a class="tips" data-tip="<?php echo esc_attr( $table_text['i18n_state_code_tip'] ); ?>">[?]</a></th>
						<th><?php echo esc_html( $table_text['i18n_zip_code'] ); ?>&nbsp;<a class="tips" data-tip="<?php echo esc_attr( $table_text['i18n_zip_code_tip'] ); ?>">[?]</a></th>
						<th class="cost"><?php echo esc_html( $table_text['i18n_line_cost'] ); ?>&nbsp;<a class="tips" data-tip="<?php echo esc_attr( $table_text['i18n_line_cost_tip'] ); ?>">[?]</a></th>
						<th class="item_cost"><?php echo esc_html( $table_text['i18n_item_cost'] ); ?>&nbsp;<a class="tips" data-tip="<?php echo esc_attr( $table_text['i18n_item_cost_tip'] ); ?>">[?]</a></th>
					</tr>
				</thead>
				<tfoot>
					<tr>
						<th colspan="6">
							<?php if ( $count < self::$rule_count_limit ) : ?>
								<a href="#" class="button button-primary insert" data-postid="<?php echo esc_attr( $post_id ); ?>"><?php echo esc_html( $table_text['i18n_insert_row'] ); ?></a>
							<?php endif; ?>

							<?php if ( $count <= self::$rule_count_limit ) : ?>
								<a href="#" class="button remove"><?php echo esc_html( $table_text['i18n_remove_row'] ); ?></a>
							<?php endif; ?>

							<a href="#" class="button export" data-postid="<?php echo esc_attr( $post_id ); ?>"><?php echo esc_html( $table_text['i18n_export_csv'] ); ?></a>
							<a href="<?php echo esc_url( $table_urls['import_url'] ); ?>" class="button import"><?php echo esc_html( $table_text['i18n_import_csv'] ); ?></a>
							<a href="<?php echo esc_url( sprintf( $table_urls['import_override_url'], absint( $post_id ) ) ); ?>" onclick="return confirm( '<?php echo esc_js( $table_text['i18n_override_conf_text'] ); ?>' );" class="button import"><?php echo esc_html( $table_text['i18n_import_csv_override'] ); ?></a>
						</th>
					</tr>
				</tfoot>
				<tbody>
					<?php
					if ( $count <= self::$rule_count_limit ) {
						$rules = self::get_shipping_rules( $post_id );

						foreach ( $rules as $rule ) {
							?>
							<tr>
								<td class="sort">&nbsp;<input type="hidden" value="<?php echo esc_attr( $rule->rule_order ); ?>" name="per_product_order[<?php echo esc_attr( $post_id ); ?>][<?php echo esc_attr( $rule->rule_id ); ?>]" /></td>
								<td class="country"><input type="text" class="wcspp_country_validation" maxlength="2" value="<?php echo esc_attr( $rule->rule_country ); ?>" placeholder="*" name="per_product_country[<?php echo esc_attr( $post_id ); ?>][<?php echo esc_attr( $rule->rule_id ); ?>]" /></td>
								<td class="state"><input type="text" maxlength="7" value="<?php echo esc_attr( $rule->rule_state ); ?>" placeholder="*" name="per_product_state[<?php echo esc_attr( $post_id ); ?>][<?php echo esc_attr( $rule->rule_id ); ?>]" /></td>
								<td class="postcode"><input type="text" value="<?php echo esc_attr( $rule->rule_postcode ); ?>" placeholder="*" name="per_product_postcode[<?php echo esc_attr( $post_id ); ?>][<?php echo esc_attr( $rule->rule_id ); ?>]" /></td>
								<td class="cost"><input type="text" class="wc_input_price input-text regular-input" value="<?php echo esc_attr( $rule->rule_cost ); ?>" placeholder="<?php echo esc_attr( $cost_placeholder ); ?>" name="per_product_cost[<?php echo esc_attr( $post_id ); ?>][<?php echo esc_attr( $rule->rule_id ); ?>]" /></td>
								<td class="item_cost"><input type="text" class="wc_input_price input-text regular-input" value="<?php echo esc_attr( $rule->rule_item_cost ); ?>" placeholder="<?php echo esc_attr( $cost_placeholder ); ?>" name="per_product_item_cost[<?php echo esc_attr( $post_id ); ?>][<?php echo esc_attr( $rule->rule_id ); ?>]" /></td>
							</tr>
							<?php
						}
					} else {
						?>
						<tr>
							<td colspan="6" style="text-align:center;padding:50px 1%;">
								<?php
								printf(
									// translators: %1$s and %2$s is html tags.
									esc_html__( '%1$sNOTICE:%2$s Only tables with %3$d rules or fewer can be edited within the product editor. %4$sPlease export your rules, modify as needed, and use the "Import CSV (override)" button to update this product\'s rules.', 'woocommerce-shipping-per-product' ),
									'<h2>',
									'</h2>',
									esc_html( self::$rule_count_limit ),
									'<br>'
								);
								?>
							</td>
						</tr>
						<?php
					}
					?>
				</tbody>
			</table>
		</div>
		<?php
	}

	/**
	 * Save.
	 *
	 * @param int $post_id Post ID.
	 */
	public function save( $post_id ) {
		// phpcs:disable WordPress.Security.NonceVerification.Missing
		// Enabled or Disabled.
		$enabled              = ! empty( $_POST['_per_product_shipping'] );
		$saved_enable_setting = get_post_meta( $post_id, '_per_product_shipping', true ) === 'yes';
		self::clear_shipping_cache_if_necessary( $enabled, $saved_enable_setting );

		if ( $enabled ) {
			update_post_meta( $post_id, '_per_product_shipping', 'yes' );
			update_post_meta( $post_id, '_per_product_shipping_add_to_all', ! empty( $_POST['_per_product_shipping_add_to_all'][ $post_id ] ) ? 'yes' : 'no' );
		} else {
			delete_post_meta( $post_id, '_per_product_shipping' );
			delete_post_meta( $post_id, '_per_product_shipping_add_to_all' );
		}

		$countries  = ! empty( $_POST['per_product_country'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_country'][ $post_id ] ) ) : array();
		$states     = ! empty( $_POST['per_product_state'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_state'][ $post_id ] ) ) : array();
		$postcodes  = ! empty( $_POST['per_product_postcode'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_postcode'][ $post_id ] ) ) : array();
		$costs      = ! empty( $_POST['per_product_cost'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_cost'][ $post_id ] ) ) : array();
		$item_costs = ! empty( $_POST['per_product_item_cost'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_item_cost'][ $post_id ] ) ) : array();
		$order      = ! empty( $_POST['per_product_order'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_order'][ $post_id ] ) ) : array();
		// phpcs:enable WordPress.Security.NonceVerification.Missing

		if ( ! empty( $countries ) ) {
			$data = compact( 'countries', 'states', 'postcodes', 'costs', 'item_costs', 'order' );
			self::save_product_rules( $post_id, $data );
		}
	}

	/**
	 * Replaces the aseterisks with emtpy string.
	 *
	 * @param string $rule Rule.
	 *
	 * @return string
	 */
	public static function replace_aseterisk( $rule ) {
		if ( ! empty( $rule ) && '*' === $rule ) {
			return '';
		}

		return $rule;
	}

	/**
	 * Save a variation.
	 *
	 * @param int $post_id ID of the variation being saved.
	 * @param int $index   Index.
	 */
	public function save_variation( $post_id, $index ) {
		// phpcs:disable WordPress.Security.NonceVerification.Missing
		$enabled              = isset( $_POST['_per_variation_shipping'][ $post_id ] );
		$countries            = ! empty( $_POST['per_product_country'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_country'][ $post_id ] ) ) : array();
		$states               = ! empty( $_POST['per_product_state'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_state'][ $post_id ] ) ) : array();
		$postcodes            = ! empty( $_POST['per_product_postcode'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_postcode'][ $post_id ] ) ) : array();
		$costs                = ! empty( $_POST['per_product_cost'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_cost'][ $post_id ] ) ) : array();
		$item_costs           = ! empty( $_POST['per_product_item_cost'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_item_cost'][ $post_id ] ) ) : array();
		$order                = ! empty( $_POST['per_product_order'][ $post_id ] ) ? wc_clean( wp_unslash( $_POST['per_product_order'][ $post_id ] ) ) : array();
		$saved_enable_setting = get_post_meta( $post_id, '_per_product_shipping', true ) === 'yes';

		self::clear_shipping_cache_if_necessary( $enabled, $saved_enable_setting );

		if ( $enabled ) {
			update_post_meta( $post_id, '_per_product_shipping', 'yes' );
			update_post_meta( $post_id, '_per_product_shipping_add_to_all', ! empty( $_POST['_per_product_shipping_add_to_all'][ $post_id ] ) ? 'yes' : 'no' );

			$data = compact( 'countries', 'states', 'postcodes', 'costs', 'item_costs', 'order' );
			self::save_product_rules( $post_id, $data );
		} else {
			delete_post_meta( $post_id, '_per_product_shipping' );
			delete_post_meta( $post_id, '_per_product_shipping_add_to_all' );
		}
		// phpcs:enable WordPress.Security.NonceVerification.Missing
	}

	/**
	 * Formatting rules data from the requests.
	 *
	 * @param array $rules Rules data.
	 *
	 * @return array Formatted rules data.
	 */
	public static function formatting_rules_data( $rules ) {
		$formatted_data = array(
			'countries'  => array(),
			'states'     => array(),
			'postcodes'  => array(),
			'costs'      => array(),
			'item_costs' => array(),
			'order'      => array(),
		);

		if ( ! is_array( $rules ) || empty( $rules ) ) {
			return $formatted_data;
		}

		$rule_data_names = array(
			'rule_country',
			'rule_state',
			'rule_postcode',
			'rule_cost',
			'rule_item_cost',
			'rule_order',
			'rule_id',
			'product_id',
		);

		foreach ( $rules as $rule ) {
			foreach ( $rule_data_names as $data_name ) {
				if ( ! isset( $rule[ $data_name ] ) ) {
					continue 2;
				}
			}

			if ( 'new' === $rule['rule_id'] ) {
				$formatted_data['countries'][ $rule['rule_id'] ][]  = $rule['rule_country'];
				$formatted_data['states'][ $rule['rule_id'] ][]     = $rule['rule_state'];
				$formatted_data['postcodes'][ $rule['rule_id'] ][]  = $rule['rule_postcode'];
				$formatted_data['costs'][ $rule['rule_id'] ][]      = $rule['rule_cost'];
				$formatted_data['item_costs'][ $rule['rule_id'] ][] = $rule['rule_item_cost'];
				$formatted_data['order'][ $rule['rule_id'] ][]      = $rule['rule_order'];
			} else {
				$formatted_data['countries'][ $rule['rule_id'] ]  = $rule['rule_country'];
				$formatted_data['states'][ $rule['rule_id'] ]     = $rule['rule_state'];
				$formatted_data['postcodes'][ $rule['rule_id'] ]  = $rule['rule_postcode'];
				$formatted_data['costs'][ $rule['rule_id'] ]      = $rule['rule_cost'];
				$formatted_data['item_costs'][ $rule['rule_id'] ] = $rule['rule_item_cost'];
				$formatted_data['order'][ $rule['rule_id'] ]      = $rule['rule_order'];
			}
		}

		return $formatted_data;
	}

	/**
	 * Get rules data for REST.
	 *
	 * @param int $product_id Product ID.
	 */
	public static function get_rest_formatted_rules( $product_id ) {
		$rules = self::get_shipping_rules( $product_id );

		if ( empty( $rules ) ) {
			return array();
		}

		$rule_count = count( $rules );
		$rule_index = array();

		for ( $i = 0; $i < $rule_count; $i++ ) {
			$rule_index[] = $i;
		}

		return array_map(
			function( $rule, $index ) {
				return array(
					'product_id'     => $rule->product_id,
					'rule_cost'      => $rule->rule_cost,
					'rule_country'   => $rule->rule_country,
					'rule_id'        => $rule->rule_id,
					'rule_item_cost' => $rule->rule_item_cost,
					'rule_order'     => $rule->rule_order,
					'rule_postcode'  => $rule->rule_postcode,
					'rule_state'     => $rule->rule_state,
					'rule_index'     => $index,
					'is_checked'     => false,
					'remove'         => false,
				);
			},
			$rules,
			$rule_index
		);
	}

	/**
	 * Save product rules.
	 *
	 * @since 2.2.9
	 * @version 2.2.9
	 *
	 * @param int   $product_id Product ID.
	 * @param array $data       Data.
	 */
	public static function save_product_rules( $product_id, $data ) {
		global $wpdb;

		$data = wp_parse_args(
			$data,
			array(
				'countries'  => array(),
				'states'     => array(),
				'postcodes'  => array(),
				'costs'      => array(),
				'item_costs' => array(),
				'order'      => array(),
			)
		);

		$countries  = $data['countries'];
		$states     = $data['states'];
		$postcodes  = $data['postcodes'];
		$costs      = $data['costs'];
		$item_costs = $data['item_costs'];
		$rule_order = $data['order'];

		foreach ( $countries as $key => $value ) {
			if ( 'new' === $key ) {
				foreach ( $value as $new_key => $new_value ) {
					$has_column_with_value = (
						! empty( $countries[ $key ][ $new_key ] )
						|| ! empty( $states[ $key ][ $new_key ] )
						|| ! empty( $postcodes[ $key ][ $new_key ] )
						|| ( isset( $costs[ $key ][ $new_key ] ) && is_numeric( $costs[ $key ][ $new_key ] ) )
						|| ( isset( $item_costs[ $key ][ $new_key ] ) && is_numeric( $item_costs[ $key ][ $new_key ] ) )
					);

					if ( $has_column_with_value ) {
						$wpdb->insert(
							$wpdb->prefix . 'woocommerce_per_product_shipping_rules',
							array(
								'rule_country'   => esc_attr( self::replace_aseterisk( $countries[ $key ][ $new_key ] ) ),
								'rule_state'     => esc_attr( self::replace_aseterisk( $states[ $key ][ $new_key ] ) ),
								'rule_postcode'  => esc_attr( self::replace_aseterisk( $postcodes[ $key ][ $new_key ] ) ),
								'rule_cost'      => esc_attr( $costs[ $key ][ $new_key ] ),
								'rule_item_cost' => esc_attr( $item_costs[ $key ][ $new_key ] ),
								'rule_order'     => absint( $rule_order[ $key ][ $new_key ] ),
								'product_id'     => absint( $product_id ),
							)
						);
					}
				}
			} else {
				$has_column_with_value = (
					! empty( $countries[ $key ] )
					|| ! empty( $states[ $key ] )
					|| ! empty( $postcodes[ $key ] )
					|| ( isset( $costs[ $key ] ) && is_numeric( $costs[ $key ] ) )
					|| ( isset( $item_costs[ $key ] ) && is_numeric( $item_costs[ $key ] ) )
				);

				if ( $has_column_with_value ) {
					$wpdb->update(
						$wpdb->prefix . 'woocommerce_per_product_shipping_rules',
						array(
							'rule_country'   => esc_attr( self::replace_aseterisk( $countries[ $key ] ) ),
							'rule_state'     => esc_attr( self::replace_aseterisk( $states[ $key ] ) ),
							'rule_postcode'  => esc_attr( self::replace_aseterisk( $postcodes[ $key ] ) ),
							'rule_cost'      => esc_attr( $costs[ $key ] ),
							'rule_item_cost' => esc_attr( $item_costs[ $key ] ),
							'rule_order'     => absint( $rule_order[ $key ] ),
						),
						array(
							'product_id' => absint( $product_id ),
							'rule_id'    => absint( $key ),
						)
					);
				} else {
					$wpdb->query( $wpdb->prepare( "DELETE FROM {$wpdb->prefix}woocommerce_per_product_shipping_rules WHERE product_id = %d AND rule_id = %s;", absint( $product_id ), absint( $key ) ) );
				}
			}
		}
	}

	/**
	 * Duplicate rules when duplicating product.
	 *
	 * @since 2.2.9
	 * @version 2.2.9
	 *
	 * @param WC_Product $duplicate Duplicated product.
	 * @param WC_Product $product   Original product.
	 */
	public function duplicate_rules( $duplicate, $product ) {
		$product   = wc_get_product( $product );
		$duplicate = wc_get_product( $duplicate );
		if ( ! $product || ! $duplicate ) {
			return;
		}

		if ( $this->is_per_product_shipping_enabled( $product ) ) {
			$this->duplicate_rules_query( $duplicate->get_id(), $product->get_id() );
		}

		// Variations may have their own rules. So, if current product is not
		// a variable, we're done duplicating.
		if ( ! $product->is_type( 'variable' ) ) {
			return;
		}

		$original_variation_ids  = $product->get_children();
		$duplicate_variation_ids = $duplicate->get_children();

		// Duplciate rules for each variation.
		foreach ( $original_variation_ids as $index => $variation_id ) {
			if ( empty( $duplicate_variation_ids[ $index ] ) ) {
				continue;
			}

			if ( $this->is_per_product_shipping_enabled( wc_get_product( $variation_id ) ) ) {
				$this->duplicate_rules_query( $duplicate_variation_ids[ $index ], $variation_id );
			}
		}
	}

	/**
	 * Checks whether a given `$product` enables per product shipping.
	 *
	 * @since 2.2.9
	 * @version 2.2.9
	 *
	 * @param WC_Product $product Product object.
	 *
	 * @return bool
	 */
	private function is_per_product_shipping_enabled( $product ) {
		return 'yes' === $product->get_meta( '_per_product_shipping' );
	}

	/**
	 * Runs query to duplicate the rules from `$product_id` to `$duplciate_id`.
	 *
	 * @since 2.2.9
	 * @version 2.2.9
	 *
	 * @param int $duplicate_id Duplicated product ID.
	 * @param int $product_id   Original product ID.
	 *
	 * @return int|false Number of rows affected/selected or false on error.
	 */
	private function duplicate_rules_query( $duplicate_id, $product_id ) {
		global $wpdb;

		return $wpdb->query(
			$wpdb->prepare(
				"
				INSERT INTO {$wpdb->prefix}woocommerce_per_product_shipping_rules
					( product_id, rule_country, rule_state, rule_postcode,
					rule_cost, rule_item_cost, rule_order )
				SELECT %d, rule_country, rule_state, rule_postcode, rule_cost,
					rule_item_cost, rule_order
				FROM {$wpdb->prefix}woocommerce_per_product_shipping_rules
				WHERE product_id = %d;
				",
				$duplicate_id,
				$product_id
			)
		);
	}

	/**
	 * Clear shipping cache if it is necessary.
	 * Do not clear when option was disabled and new is also disable.
	 * Any other needs clearing.
	 *
	 * @since 2.2.16
	 * @version 2.2.16
	 *
	 * @param int $enabled is per-product enabled in this save action.
	 * @param int $saved_enable_setting per-product saved value.
	 */
	public static function clear_shipping_cache_if_necessary( $enabled, $saved_enable_setting ) {
		global $wpdb;

		// If per product was disabled and it still is we don't need to clear cache.
		if ( false === $saved_enable_setting && false === $enabled ) {
			return;
		}

		// Increments the transient version to invalidate cache.
		WC_Cache_Helper::get_transient_version( 'shipping', true );
	}

	/**
	 * Retrieve and send shipping rules with the given product ID.
	 */
	public function export_rules() {
		check_ajax_referer( 'per-product-export', 'security' );

		if ( ! current_user_can( 'edit_products' ) ) {
			wp_send_json_error( array( 'message' => __( 'Insufficient privileges to export product shipping rules.', 'woocommerce-shipping-per-product' ) ) );
		}

		$product_id = isset( $_POST['product_id'] ) ? absint( $_POST['product_id'] ) : 0;
		$response   = array(
			'success' => false,
		);

		// Make sure the product ID was passed.
		if ( empty( $product_id ) ) {
			wp_send_json_error( array( 'message' => __( 'Invalid product ID.', 'woocommerce-shipping-per-product' ) ) );
		}

		global $wpdb;
		$response['rules'] = $wpdb->get_results( $wpdb->prepare( "SELECT * FROM {$wpdb->prefix}woocommerce_per_product_shipping_rules WHERE product_id = %d ORDER BY rule_order;", $product_id ) );

		// If we have results, change 'success' to true.
		if ( ! empty( $response['rules'] ) ) {
			$response['success'] = true;
		}

		wp_send_json( $response );
	}
}
